-- | Maintainer: 2016 Evan Cofsky <evan@theunixman.com>
-- 
-- ZFS properties

module Propellor.Property.ZFS (
	module Propellor.Property.ZFS.Properties,
	module Propellor.Types.ZFS
) where

import Propellor.Property.ZFS.Properties
import Propellor.Types.ZFS
